import os
import httpx

from datetime import datetime
from ..utils import downloader
from ..utils.custom_types import UrlDictionary


class Common():

    def common_filename_creator(self, url_dictionary: UrlDictionary) -> str:
        '''
        Builds out the general standard filename format. It adds receives the information from
        url_dictionary and adds it to the filename string where appropriate. The standard format is:

        prefix-date_uploaded-album_name-count-id-description.ext

        prefix - name of the website
        date_uploaded - if a datetime from when it is uploaded is provided it is added
        album - add the album name the file belongs to (if it applies)
        count - applies if the file is apart of an album. Tries to keep the files in order
        name - name of the file. usually the id from the the url
        description - usually the title given to the file from the orignal site it was uploaded
        to. typically a sentence
        extension - the extension of the file
        '''
        # Template of all information that should be passed
        # in the url_dictionary. Copy and past into the extractor

        # if 'prefix' in url_dictionary:
        #     prefix = url_dictionary['prefix']
        # else:
        #     prefix = self.prefix
        #
        # if 'album_name' in url_dictionary:
        #     album_name = url_dictionary['album_name']
        # else:
        #     album_name = None
        #
        # if 'count' in url_dictionary:
        #     count = url_dictionary['count']
        # else:
        #     count = None
        #
        # if 'description' in url_dictionary:
        #     description = url_dictionary['description']
        # else:
        #     description = None

        # url_dictionary['name'] = name
        # url_dictionary['date'] = date
        # url_dictionary['url_to_download'] = url_to_download
        # url_dictionary['url_to_record'] = url_to_record
        # url_dictionary['file_size'] = file_size
        # url_dictionary['extension'] = extension
        # url_dictionary['filename'] = self.common_filename_creator(url_dictionary)

        filename = ''

        if url_dictionary.get('prefix'):
            filename += f"{url_dictionary['prefix']}-"

        if url_dictionary.get('date'):
            dt = datetime.fromtimestamp(
                url_dictionary['date']).strftime('%Y-%m-%d')
            filename += f"{dt}-"

        if url_dictionary.get('album_name'):
            filename += f"{url_dictionary['album_name']}-"

        if url_dictionary.get('count'):
            filename += f"{url_dictionary['count']:03}-"

        if url_dictionary.get('name'):
            filename += f"{url_dictionary['name']}"

        if url_dictionary.get('description'):
            filename += f"-{url_dictionary['description']}"

        extension = self.common_find_extension(
            url_dictionary['extension']).lower()

        filename = filename.replace('\n', '  ').replace('\r', '  ')

        invalid_characters = ['/', '\\']
        filename = ''.join(
            c for c in filename if c not in invalid_characters)

        if len(str(filename)) > 220:
            filename = filename[:220]

        filename += f".{extension.lower()}"

        return filename

    async def common_file_downloader(self, url_dictionary: UrlDictionary) -> None:

        if url_dictionary.get("download"):

            await downloader.download_file(url_dictionary)

        else:

            if url_dictionary.get('url_to_record'):

                print(url_dictionary['url_to_record'])

            else:
                print(url_dictionary['url_to_download'])

    async def common_generic_downloader(self, url: str) -> None:

        response = httpx.get(url, verify=False)

        # print(response.headers)
        file_size: int = int(response.headers.get('content-length', 1))
        filename: str = url.split('/')[-1]

        url_dictionary: UrlDictionary = {}
        url_dictionary['url_to_download'] = url
        url_dictionary['filename'] = filename
        url_dictionary['file_size'] = file_size
        url_dictionary['response'] = response

        await downloader.download_file(url_dictionary.copy())

    def common_find_extension(self, extension: str) -> str:

        if extension in MIME_TYPES:
            return MIME_TYPES[extension]

        elif extension in MIME_TYPES.values():
            return extension

        else:
            return '???'


MIME_TYPES = {
    "image/jpeg": "jpg",
    "image/jpg": "jpg",
    "image/png": "png",
    "image/gif": "gif",
    "image/bmp": "bmp",
    "image/x-bmp": "bmp",
    "image/x-ms-bmp": "bmp",
    "image/webp": "webp",
    "image/avif": "avif",
    "image/svg+xml": "svg",
    "image/ico": "ico",
    "image/icon": "ico",
    "image/x-icon": "ico",
    "image/vnd.microsoft.icon": "ico",
    "image/x-photoshop": "psd",
    "application/x-photoshop": "psd",
    "image/vnd.adobe.photoshop": "psd",

    "video/webm": "webm",
    "video/ogg": "ogg",
    "video/mp4": "mp4",
    "video/x-m4v": "m4v",
    "video/x-matroska": "mkv",
    "video/x-ms-asf": "wmv",
    "video/x-msvideo": "avi",
    "video/x-flv": "flv",
    "video/quicktime": "mov",
    "video/x-wav": "wav",

    "audio/wav": "wav",
    "audio/x-wav": "wav",
    "audio/webm": "webm",
    "audio/ogg": "ogg",
    "audio/mpeg": "mp3",
    "audio/x-m4a": "m4a",
    "audio/mpeg": "mp3",

    "application/zip": "zip",
    "application/x-zip": "zip",
    "application/x-zip-compressed": "zip",
    "application/rar": "rar",
    "application/x-rar": "rar",
    "application/x-rar-compressed": "rar",
    "application/x-7z-compressed": "7z",

    "application/pdf": "pdf",
    "application/x-pdf": "pdf",
    "application/x-shockwave-flash": "swf",

    "application/ogg": "ogg",
    # https://www.iana.org/assignments/media-types/model/obj
    "model/obj": "obj",
    "application/octet-stream": "bin",
}



from typing_extensions import NotRequired, TypedDict
import httpx


class UrlDictionary(TypedDict):

    album_name: NotRequired[str]
    cookies: NotRequired[dict[str, str]]
    count: NotRequired[int]
    date: NotRequired[str | float]
    description: NotRequired[str]
    download: NotRequired[bool]
    extension: NotRequired[str]
    filename: NotRequired[str]
    file_size: NotRequired[int | None]
    name: NotRequired[str]
    prefix: NotRequired[str]
    reddit_uniq_id: NotRequired[str]
    reddit_username: NotRequired[str]
    response: NotRequired[httpx.Response]
    retries: NotRequired[int]
    url: NotRequired[str]
    url_to_download: NotRequired[str]
    url_to_record: NotRequired[str]
    ytdlp_required: NotRequired[bool]



import hashlib
import logging
import os

from . import rat_info
from multiprocessing.pool import ThreadPool
from pathlib import Path

log = logging.getLogger(__name__)


async def file_hasher():

    # temp = Path().cwd().parent
    # path_to_folder = Path(temp).parent
    # folder_name = Path(temp).name
    # print(path_to_folder)
    # print(folder_name)
    # full_path = PurePath(path_to_folder, folder_name)
    # print(full_path)

    log.info("Searching for matching files")
    files_to_delete = await finding_duplicates()

    log.info("Deleting duplicate files")
    try:
        for root, dirs, files in os.walk(os.getcwd()):
            for file in files_to_delete:
                if file in files:
                    Path(Path(file).resolve()).unlink()

    except FileNotFoundError:
        pass

    log.info(f"Removed {len(files_to_delete)} files")


async def finding_duplicates():

    files = [file for file in Path().cwd().iterdir() if file.is_file()]
    files_to_delete = []
    file_hashes = rat_info.get_file_hashes()

    no_rat_file = False

    if file_hashes is None:
        no_rat_file = True
        file_hashes = {}

    log.info("Calculating hashes")
    number_of_workers = os.cpu_count()
    with ThreadPool(number_of_workers) as pool:
        files_hash = pool.map(file_hash, files)

    for file in files_hash:

        hash = file[0]
        name = Path(file[1]).name

        if name == rat_info.get_rat_name():
            continue

        elif hash not in file_hashes:
            file_hashes[hash] = name

        if len(file_hashes[hash]) < len(name):
            files_to_delete.append(file_hashes[hash])
            file_hashes[hash] = name

        elif len(file_hashes[hash]) > len(name):
            files_to_delete.append(name)

        elif len(file_hashes[hash]) == len(name):

            current_file_dashes = file_hashes[hash].count('-')
            new_file_dashes = name.count('-')

            if current_file_dashes < new_file_dashes:
                files_to_delete.append(file_hashes[hash])
                file_hashes[hash] = name

            elif current_file_dashes > new_file_dashes:
                files_to_delete.append(name)

            elif current_file_dashes == new_file_dashes and file_hashes[hash] != name:
                files_to_delete.append(name)

    log.info("Returning found hashes")

    if no_rat_file is not True:
        rat_info.update_rat(category_1='file_hashes', category_2=file_hashes)
    return files_to_delete


def file_hash(file):

    buffer_size = 1024 * 1024

    with open(file, 'rb') as file:

        md5 = hashlib.md5()
        while True:
            data = file.read(buffer_size)
            if not data:
                break
            md5.update(data)

        hash = md5.hexdigest()
        return (hash, file.name)



import logging
import re

from ..extractors.bunkr import Bunkr
from ..extractors.cyberdrop import Cyberdrop
from ..extractors.imgur import Imgur
from ..extractors.jpg import Jpg
from ..extractors.gofile import Gofile
from ..extractors.gfycat import Gfycat
from ..extractors.reddit import Reddit
from ..extractors.redgifs import Redgifs
from .custom_types import UrlDictionary

log = logging.getLogger(__name__)


stored_class_instances = dict()


bunkr_re = re.compile(r"(https?://)([\w\-]+\.)?(bunkr\.(ru|la|su|is))")

cyberdrop_re = re.compile(
    r"(https?://)(www\.|[\w\-]+\.)?(cyberdrop\.(me|cc|to))")

gofile_re = re.compile(r"(https?://)(gofile\.io)")

gfycat_re = re.compile(r"(https?://)(gfycat\.com(/))")

imgur_re = re.compile(r"(https?://)?(www\.|[im]\.)?imgur\.(?:com|io)")

jpg_re = re.compile(r"(https?://)(\w+\.)?(jpg\.(church|fish))")

reddit_re = re.compile(r"(https?://)([iv]\.|www\.)?(reddit|redd\.it)")

redgifs_re = re.compile(r"(https?://)(v3\.|www\.|i.)?(redgifs.com)")


async def validate_url(url_dictionary: UrlDictionary) -> None:

    # elif _re.search(url_dictionary['url']):
    #     log.info('[] Attempting Extraction')
    #     await (url_dictionary)
    #     return

    if bunkr_re.search(url_dictionary['url']):
        log.info('[BUNKR] Attempting Extraction')
        await bunkr(url_dictionary)
        return

    elif cyberdrop_re.search(url_dictionary['url']):
        log.info('[CYBERDROP] Attempting Extraction')
        await cyberdrop(url_dictionary)
        return

    elif gofile_re.search(url_dictionary['url']):
        log.info('[GOFILE] Attempting Extraction')
        await gofile(url_dictionary)
        return

    elif gfycat_re.search(url_dictionary['url']):
        log.info('[GFYCAT] Attempting Extraction')
        await gfycat(url_dictionary)
        return

    elif imgur_re.search(url_dictionary['url']):
        log.info('[IMGUR] Attempting Extraction')
        await imgur(url_dictionary)
        return

    elif jpg_re.search(url_dictionary['url']):
        log.info('[JPG] Attempting Extraction')
        await jpg(url_dictionary)
        return

    elif redgifs_re.search(url_dictionary['url']):
        log.info('[REDGIFS] Attempting Extraction')
        await redgifs(url_dictionary)
        return

    elif reddit_re.search(url_dictionary['url']):
        log.info('[REDDIT] Attempting Extraction')
        await reddit(url_dictionary)
        return

    else:
        log.info(f"No extractor found for: {url_dictionary['url']}")


async def template(url_dictionary: UrlDictionary) -> None:

    if "" not in stored_class_instances.keys():
        log.debug("Creating  class")
        stored_class_instances[""] = Imgur()

    log.debug("Passing to ")
    await stored_class_instances[""].run(url_dictionary)


async def bunkr(url_dictionary: UrlDictionary) -> None:

    if "bunkr" not in stored_class_instances.keys():
        log.debug("Creating bunkr class")
        stored_class_instances["bunkr"] = Bunkr()

    log.debug("Passing to bunkr")
    await stored_class_instances["bunkr"].run(url_dictionary)


async def cyberdrop(url_dictionary: UrlDictionary) -> None:

    if "cyberdrop" not in stored_class_instances.keys():
        log.debug("Creating cyberdrop class")
        stored_class_instances["cyberdrop"] = Cyberdrop()

    log.debug("Passing to cyberdrop")
    await stored_class_instances["cyberdrop"].run(url_dictionary)


async def gofile(url_dictionary: UrlDictionary) -> None:

    if "gofile" not in stored_class_instances.keys():
        log.debug("Creating gofile class")
        stored_class_instances["gofile"] = Gofile()

    log.debug("Passing to gofile")
    await stored_class_instances["gofile"].run(url_dictionary)


async def gfycat(url_dictionary: UrlDictionary) -> None:

    if "gfycat" not in stored_class_instances.keys():
        log.debug("Creating gfycat class")
        stored_class_instances["gfycat"] = Gfycat()

    log.debug("Passing to gfycat")
    await stored_class_instances["gfycat"].run(url_dictionary)


async def imgur(url_dictionary: UrlDictionary) -> None:

    if "imgur" not in stored_class_instances.keys():
        log.debug("Creating imgur class")
        stored_class_instances["imgur"] = Imgur()

    log.debug("Passing to imgur")
    await stored_class_instances["imgur"].run(url_dictionary)


async def jpg(url_dictionary: UrlDictionary) -> None:

    if "jpg" not in stored_class_instances.keys():
        log.debug("Creating jpg class")
        stored_class_instances["jpg"] = Jpg()

    log.debug("Passing to jpg")
    await stored_class_instances["jpg"].run(url_dictionary)


async def reddit(url_dictionary: UrlDictionary) -> None:

    if "reddit" not in stored_class_instances.keys():
        log.debug("Creating reddit class")
        stored_class_instances["reddit"] = Reddit()

    log.debug("Passing to reddit")
    await stored_class_instances["reddit"].run(url_dictionary)


async def redgifs(url_dictionary: UrlDictionary) -> None:

    if "redgifs" not in stored_class_instances.keys():
        log.debug("Creating redgifs class")
        stored_class_instances["redgifs"] = Redgifs()

    log.debug("Passing to redgifs")
    await stored_class_instances["redgifs"].run(url_dictionary)



import logging
import sys


def create_logger(level=50):

    logging.basicConfig(level=level,
                        stream=sys.stdout,
                        format="%(asctime)s — %(name)s — %(levelname)s — %(funcName)s:%(lineno)d — %(message)s",
                        datefmt='%Y/%m/%d %H:%M:%S'
                        )
    httpx = logging.getLogger('httpx')
    httpx.setLevel(logging.WARNING)



import argparse


def make_args():
    parser = argparse.ArgumentParser(
        prog='ripandtear', description='A file downloader/archive manager')

    general = parser.add_argument_group('general', 'General Options')
    general.add_argument(
        '-v', '--version', help='prints out the current version', action='store_true')
    general.add_argument('-gd', '--generic-download',
                         help="Generic downloader that doesn't store url", metavar='', action='append')
    general.add_argument(
        '-g', '--get-urls', help='Print URLs instead of downloading', metavar='', action='append')
    general.add_argument('-l', '--log-level',
                         help='Set log level to print to the screen. 1 = debug, 2 = info, 3 = warning, 4 = error, 5 = critical. Ex: -l 1, -l 2', metavar='', action='store')
    general.add_argument('-S', '--sort-files',
                         help='Sort files into pics/vids/audio/text directories', action='store_true')
    general.add_argument('-H', '--hash-files',
                         help='Finds the hashes of files and deletes duplicates. If no .rat exists then it will hash the local files. If a .rat exists \
                                 it will compare the file hash with the stored hashes in the .rat and remove the matching file \
                                 with the shorter file name', action='store_true')
    general.add_argument('-ee', '--erase-errors',
                         help='Erase the error directories that are stored in the current .rat file', action='store_true')

    adding = parser.add_argument_group('adding', 'Adds Information to .rat')
    adding.add_argument(
        '-c', '--coomer', help='adds a coomer.party url to the .rat file', metavar='', action='append')
    adding.add_argument('-cb', '--chaturbate',
                        help='adds chaturbate username to .rat file. Can add multiple names if seperated by a comma (,). Ex: name,name1', metavar='', action='append')
    adding.add_argument('-d', '--download',
                        help='give a url to download content from a supported website', metavar='', action='append')
    adding.add_argument('-f', '--fansly', help='adds fansly username to .rat file. Can add multiple names if seperated by a comma (,). Ex: name,name1',
                        metavar='', type=str, action='append')
    adding.add_argument('-i', '--instagram', help='adds instagram username to .rat file. Can add multiple names if seperated by a comma (,). Ex: name,name1',
                        metavar='', type=str, action='append')
    adding.add_argument('-mfc', '--myfreecams', help='adds myfreecams username to .rat file. Can add multiple names if seperated by a comma (,). Ex: name,name1',
                        metavar='', type=str, action='append')
    adding.add_argument('-o', '--onlyfans', help='adds onlyfans username to .rat file. Can add multiple names if seperated by a comma (,). Ex: name,name1',
                        metavar='', type=str, action='append')
    adding.add_argument('-O', '--output', help='give a desired name for a downloaded file when combined with -d',
                        metavar='', type=str, action='append')
    adding.add_argument('-p', '--pornhub', help='adds pornhub username to .rat file. Can add multiple names if seperated by a comma (,). Ex: name,name1',
                        metavar='', type=str, action='append')
    adding.add_argument('-P', '--patreon', help='adds patreon username to .rat file. Can add multiple names if seperated by a comma (,). Ex: name,name1',
                        metavar='', type=str, action='append')
    adding.add_argument('-r', '--reddit', help='adds reddit username to .rat file. Can add multiple names if seperated by a comma (,). Ex: name,name1',
                        metavar='', type=str, action='append')
    adding.add_argument('-R', '--redgifs', help='adds redgifs username to .rat file. Can add multiple names if seperated by a comma (,). Ex: name,name1',
                        metavar='', type=str, action='append')
    adding.add_argument(
        '-s', '--simp', help='adds a simpcity url to the .rat file', metavar='', action='append')
    adding.add_argument('-t', '--twitter', help='adds twitter username to .rat file. Can add multiple names if seperated by a comma (,). Ex: name,name1',
                        metavar='', type=str, action='append')
    adding.add_argument('-T', '--tiktok', help='adds tiktok username to .rat file. Can add multiple names if seperated by a comma (,). Ex: name,name1',
                        metavar='', type=str, action='append')
    adding.add_argument('-tum', '--tumblr', help='adds tumblr username to .rat file. Can add multiple names if seperated by a comma (,). Ex: name,name1',
                        metavar='', type=str, action='append')
    adding.add_argument('-twitch', '--twitch', help='adds twitch username to .rat file. Can add multiple names if seperated by a comma (,). Ex: name,name1',
                        metavar='', type=str, action='append')
    adding.add_argument('-u', '--url-add', help='saves a url to be downloaded later',
                        metavar='', type=str, action='append')
    adding.add_argument('-U', '--urls-downloaded',
                        help='add a url that has already been downloaded', metavar='', type=str, action='append')
    adding.add_argument('-y', '--youtube', help='adds youtube username to .rat file. Can add multiple names if seperated by a comma (,). Ex: name,name1',
                        metavar='', type=str, action='append')

    printing = parser.add_argument_group(
        'print', 'Prints out information from .rat file')
    printing.add_argument('-pc', '--print-coomer',
                          help='Prints out coomer.party urls', action='store_true')
    printing.add_argument('-pcb', '--print-chaturbate',
                          help='Prints out chaturbate names', action='store_true')
    printing.add_argument('-pe', '--print-errors',
                          help='Prints out stored errors', action='store_true')
    printing.add_argument('-pf', '--print-fansly',
                          help='Prints out fansly names', action='store_true')
    printing.add_argument('-pi', '--print-instagram',
                          help='Prints out instagram names', action='store_true')
    printing.add_argument('-pmfc', '--print-myfreecams',
                          help='Prints out myfreecam names', action='store_true')
    printing.add_argument('-po', '--print-onlyfans',
                          help='Prints out onlyfan names', action='store_true')
    printing.add_argument('-pp', '--print-pornhub',
                          help='Prints out pornhub names', action='store_true')
    printing.add_argument('-pP', '--print-patreon',
                          help='Prints out patreon names', action='store_true')
    printing.add_argument('-pr', '--print-reddit',
                          help='Prints out reddit names', action='store_true')
    printing.add_argument('-pR', '--print-redgifs',
                          help='Prints out redgifs names', action='store_true')
    printing.add_argument('-ps', '--print-simpcity',
                          help='Prints out simpcity urls', action='store_true')
    printing.add_argument('-pt', '--print-twitter',
                          help='Prints out twitter names', action='store_true')
    printing.add_argument('-pT', '--print-tiktok',
                          help='Prints out tiktok names', action='store_true')
    printing.add_argument('-ptum', '--print-tumblr',
                          help='Prints out tumblr names', action='store_true')
    printing.add_argument('-ptwitch', '--print-twitch',
                          help='Prints out twitch names', action='store_true')
    printing.add_argument('-pu', '--print-urls_to_download',
                          help='Prints out urls to download', action='store_true')
    printing.add_argument('-pU', '--print-urls_downloaded',
                          help='Prints out urls already downloaded', action='store_true')
    printing.add_argument('-py', '--print-youtube',
                          help='Prints out youtube names', action='store_true')

    syncing = parser.add_argument_group(
        'sync', 'Downloads the media from the appropriate user pages. Uses the stored usernames from each category to do so')
    syncing.add_argument('-sa', '--sync-all',
                         help='Syncs all availible user profiles', action='store_true')
    syncing.add_argument('-se', '--sync-errors',
                         help='Syncs the stored error directories in the .rat file', action='store_true')
    syncing.add_argument('-sr', '--sync-reddit',
                         help='Syncs Reddit users profile', action='store_true')
    syncing.add_argument('-sR', '--sync-redgifs',
                         help='Syncs Redgifs user profile', action='store_true')
    syncing.add_argument('-su', '--sync-urls-to-download',
                         help='Downloads urls in the urls to download category within a .rat (urls can be added here with -u)', action='store_true')

    return parser



import asyncio
import hashlib
from typing import Optional
import httpx
import logging
import re

from .common import Common
from ..utils import rat_info
from ..utils.custom_types import UrlDictionary
from ..utils import color

log = logging.getLogger(__name__)

re_foler_with_password = re.compile(
    r"(https?://)(gofile\.io)(/d/)(\w+|\w{8}-\w{4}-\w{4}-\w{4}-\w{12})~(\w+)")

re_folder = re.compile(
    r"(https?://)(gofile\.io)(/d/)(\w+|\w{8}-\w{4}-\w{4}-\w{4}-\w{12})")


class Gofile(Common):

    def __init__(self):
        log.debug("Getting api token")
        self.token = self.get_token()
        # print(self.token)
        self.website_token = '12345'

    async def run(self, url_dictionary: UrlDictionary) -> None:

        if re_foler_with_password.match(url_dictionary['url']):
            log.debug("url matches folder with password regex")
            temp = url_dictionary['url']
            temp = temp.split("~")
            url_dictionary['url'] = temp[0]
            content_id = re_folder.match(temp[0]).group(4)
            password = temp[1]
            await self.get_content(content_id, url_dictionary.copy(), password)

        elif re_folder.match(url_dictionary['url']):
            log.debug("url matches folder regex")
            content_id = re_folder.match(url_dictionary['url']).group(4)
            await self.get_content(content_id, url_dictionary.copy())

        else:
            log.info("Url does not match regex: {url_dictionary['url']}")

    async def call(self, endpoint, params=None) -> httpx.Response:

        log.debug("Calling api to get folder information")
        url = f"https://api.gofile.io/{endpoint}"

        async with httpx.AsyncClient() as client:

            response = await client.get(url, params=params, timeout=None)
            return response.json()

    async def get_content(self, contentId: str, url_dictionary: UrlDictionary, password: Optional[str] = None) -> None:

        already_downloaded_urls: list[str] = rat_info.get_downloaded_urls()
        if already_downloaded_urls:
            if url_dictionary['url'] in already_downloaded_urls:
                log.info(f"Already downloaded: {url_dictionary['url']}")
                color.output('blue', f"{url_dictionary['url']}")
                return

        params = {"contentId": contentId,
                  "token": self.token,
                  "websiteToken": self.website_token}

        if password:
            log.debug("password found. Hashing")
            hash = hashlib.sha256()
            hash.update(password.encode('utf8'))
            hashed_password = hash.hexdigest()
            params['password'] = hashed_password

        log.debug("Calling api")
        data = await self.call("getContent", params)

        if data['status'] != 'ok':
            log.info(f"Error downloading content: {url_dictionary['url']}")

        log.debug("Received folder information. Finding content...")
        data = data['data']
        # print(data)
        # return

        tasks = []
        content = []
        try:
            for child in data['childs']:
                content.append(child)

        except KeyError:
            log.info(
                f"File deleted. Adding to completed urls to avoid reattempt: {url_dictionary['url']}")
            color.output('red', f"{url_dictionary['url']}")
            rat_info.add_entry(category_1='urls_downloaded',
                               entry=url_dictionary['url'])
            return

        for id in content:
            item = data['contents'][id]
            if item['type'] == 'file':
                url_dictionary['name'] = item['name'].split(".")[0]
                url_dictionary['url_to_download'] = item['link']
                url_dictionary['cookies'] = {'accountToken': str(self.token)}
                url_dictionary['file_size'] = item['size']
                url_dictionary['extension'] = item['mimetype']
                url_dictionary['filename'] = self.common_filename_creator(
                    url_dictionary.copy())
                tasks.append(asyncio.create_task(
                    self.common_file_downloader(url_dictionary.copy())))

            elif item['type'] == 'folder':
                url = f"https://gofile.io/d/{id}"
                url_dictionary['url'] = url
                await self.run(url_dictionary.copy())

            else:
                log.warn("Entry is not a file or folder. Unrecognized type")
                color.output('red', f"{url_dictionary['url']}")

        await asyncio.gather(*tasks)

    def get_token(self) -> httpx.Response:

        url = "https://api.gofile.io/createAccount"

        response = httpx.get(url, timeout=None).json()
        return response['data']['token']



import aiofiles
import httpx
import logging
import os

from . import color
from . import rat_info
import yt_dlp

from pathlib import Path
from .custom_types import UrlDictionary
from tqdm import tqdm

log = logging.getLogger(__name__)


async def download_file(url_dictionary: UrlDictionary) -> None:

    log.debug("Inside downloader")

    url_to_download = str(url_dictionary['url_to_download'])
    url_dictionary['url_to_download'] = url_to_download

    if url_dictionary.get("file_size"):
        file_size = int(url_dictionary['file_size'])

    else:
        file_size = 0

    log.debug("Checking if url has already been downloaded")
    if url_dictionary.get('url_to_record'):

        if rat_info.check_existence(category_1='urls_downloaded', entry=url_dictionary['url_to_record']):
            log.info(
                f"Url already recorded: {url_dictionary['url_to_record']}")
            color.output("blue", f"\n{url_dictionary['url_to_record']}")
            return
    else:

        if rat_info.check_existence(category_1='urls_downloaded', entry=url_dictionary['url_to_download']):
            log.info(f"Url already recorded: {url_to_download}")
            color.output("blue", f"\n{url_dictionary['url_to_download']}")
            return

    log.debug("url has not been downloaded yet")

    if url_dictionary.get('ytdlp_required'):
        ytdlp_download(url_dictionary)
        return

    # is_file_local returns a tuple that says if the file exists in the
    # first position and many bytes have been downloaded in the second
    # Ex: (True, 8473)

    log.debug("Checking if file exists locally")
    filename = url_dictionary['filename']
    is_file_local = local_file_exists(filename, file_size)

    try:
        if is_file_local[0] == "True":

            if is_file_local[1] == "done":

                color.output("cyan", filename)
                return

            else:

                mode = 'ab'
                initial_position = is_file_local[1]
        else:

            mode = 'wb'
            initial_position = 0

    except TypeError:
        log.error("Filename too long. Skipping")
        color.output('red', f"\n{url_dictionary['url']}")
        rat_info.add_entry(category_1='urls_downloaded',
                           entry=str(url_dictionary['url']))
        return

    headers = {'Range': f'bytes={initial_position}-'}

    if url_dictionary.get('headers'):

        for key, value in url_dictionary['headers'].items():
            headers[key] = value

    if url_dictionary.get('response') is None:
        try:
            log.debug("Sending request")

            cookies = None
            if url_dictionary.get('cookies'):
                cookies = url_dictionary['cookies']

            async with httpx.AsyncClient() as client:
                response = await client.get(
                    f"{url_to_download}", headers=headers, timeout=None, cookies=cookies)

        except Exception as e:
            log.exception(e)
            return
    else:
        response = url_dictionary['response']

    log.debug(f"Status Code: {response.status_code}")

    if response.status_code >= 300:
        errors(response, url_dictionary)
        return

    if file_size == 0:
        file_size = response.headers.get('content-length')

    try:

        log.debug("Preparing to download")
        async with aiofiles.open(filename, mode) as file:
            bar = tqdm(bar_format='{l_bar}{bar:40}|{n_fmt}/{total_fmt}{postfix}',
                       initial=int(initial_position),
                       total=int(file_size),
                       unit='B',
                       unit_scale=True,
                       unit_divisor=1024,
                       colour='green')

            for chunk in response.iter_bytes(chunk_size=8192):
                size = await file.write(chunk)
                bar.update(size)
                bar.set_postfix_str(f"{filename}")

    except FileNotFoundError:
        log.exception(
            f"Unable to download: {url_dictionary['url_to_download']}")
        color.output('red', f"\n{url_dictionary['url_to_download']}")
        return

    try:
        if url_dictionary.get('url_to_record'):
            rat_info.add_entry(category_1='urls_downloaded',
                               entry=url_dictionary['url_to_record'])
        else:
            rat_info.add_entry(category_1='urls_downloaded',
                               entry=url_dictionary['url_to_download'])

    except ValueError as identifier:
        log.info(identifier)


def errors(response, url_dictionary):
    if response.status_code == 403:
        log.warn(
            f"403 Authorization Error: {url_dictionary['url_to_download']}")
        color.output('red', f"\n{url_dictionary['url_to_download']}")

    elif response.status_code == 404:
        log.warn(
            f"404 Not Found: {url_dictionary['url_to_download']}")
        color.output('red', f"\n{url_dictionary['url_to_download']}")
        rat_info.add_entry(category_1='urls_downloaded',
                           entry=url_dictionary['url_to_download'])

    elif response.status_code == 429:
        log.warn(f"{response.status_code}: Too Many Requests")
        rat_info.add_error_dictionary(url_dictionary)
        color.output('red', f"\n{url_dictionary['url_to_download']}")

    elif 500 <= response.status_code < 600:
        log.warn(
            f"500 error. Saving the url to be attempted later: {response.status_code} - {url_dictionary['url_to_download']}")
        rat_info.add_error_dictionary(url_dictionary)
        color.output('red', f"\n{url_dictionary['url_to_download']}")


def local_file_exists(filename, online_file_size):

    try:
        if Path(filename).exists():

            temp = open(filename)
            temp.seek(0, os.SEEK_END)
            local_file_size = temp.tell()

            if online_file_size > 0:

                if local_file_size != online_file_size:
                    return ("True", local_file_size)

                else:
                    return ("True", "done")

            else:
                return ("False")
        else:
            return ("False")

    except OSError:
        return


def ytdlp_download(url_dictionary):

    try:
        log.info("Attempting yt_dlp download")

        ytdlp_options = {"outtmpl": url_dictionary['filename']}

        with yt_dlp.YoutubeDL(ytdlp_options) as ytdlp:
            ytdlp.download(url_dictionary['url_to_download'])

        rat_info.add_entry(category_1='urls_downloaded',
                           entry=url_dictionary['url_to_download'])
        return

    except yt_dlp.utils.DownloadError:
        log.info(
            f"Problem downloading url. Saving for later: {url_dictionary['url_to_download']}")
        rat_info.add_error_dictionary(url_dictionary)
        return



import asyncio
import httpx
import logging
import re

from bs4 import BeautifulSoup
from .common import Common
from ..utils.custom_types import UrlDictionary
from ..utils import color
from ..utils import rat_info

log = logging.getLogger(__name__)

sem = asyncio.Semaphore(6)

re_cyberdrop_album = re.compile(
    r"(https?://)(www\.)?(cyberdrop\.(me))(/a/)(\w+)")

re_cyberdrop_single = re.compile(
    r"(https?://)([\w-]+\.)(cyberdrop\.(cc|to))/([\w\-_\(\)\%\+]+)\.(\w+)")


class Cyberdrop(Common):

    async def run(self, url_dictionary: UrlDictionary) -> None:

        if re_cyberdrop_single.match(url_dictionary['url']):
            await self.single_downloader(url_dictionary)

        if re_cyberdrop_album.match(url_dictionary['url']):
            await self.album_downloader(url_dictionary)

    async def single_downloader(self, url_dictionary: UrlDictionary) -> None:

        async with sem:
            log.info(
                f"Individual file found. Attempting download: {url_dictionary['url']}")
            try:
                async with httpx.AsyncClient(verify=False) as client:
                    response = await client.get(url_dictionary['url'], timeout=None, follow_redirects=True)
            except httpx.ReadError:
                log.warn(
                    f"Problem downloading url. Saving for later: {url_dictionary['url']}")
                rat_info.add_error_dictionary(url_dictionary.copy())
                return

            # print(response.headers)
            url_dictionary['response'] = response
            url_dictionary['name'] = re_cyberdrop_single.match(
                url_dictionary['url']).group(5)
            url_dictionary['extension'] = response.headers.get('content-type')
            url_dictionary['url_to_download'] = url_dictionary['url']
            url_dictionary['file_size'] = int(
                response.headers.get('content-length'))
            url_dictionary['filename'] = self.common_filename_creator(
                url_dictionary.copy())

            # print(url_dictionary)
            await self.common_file_downloader(url_dictionary.copy())

    async def album_downloader(self, url_dictionary: UrlDictionary) -> None:

        log.info(
            f"Cyberdrop album found. Attempting download: {url_dictionary['url']}")
        async with httpx.AsyncClient() as client:
            response = await client.get(url_dictionary['url'], timeout=None)

        log.debug("Got response. Parsing for individual file urls")
        try:
            soup = BeautifulSoup(response.text, 'html.parser')

            table = soup.find('div', {"id": "table"})
            hrefs = table.find_all('a', {"class": "image"})

        except AttributeError:
            log.info(f"Album deleted: {url_dictionary['url']}")
            color.output('red', f"{url_dictionary['url']}")
            rat_info.add_entry(category_1='urls_downloaded',
                               entry=url_dictionary['url'])
            return

        tasks = []
        already_downloaded_urls = rat_info.get_downloaded_urls()

        log.info("Preparing to download individual files")
        for link in hrefs:
            url = str(link['href'])
            url = url.replace(" ", "%20")

            if already_downloaded_urls is not None:
                if url in already_downloaded_urls and url_dictionary['download'] is True:
                    continue

                elif url not in already_downloaded_urls and url_dictionary['download'] is True:
                    url_dictionary['url'] = url
                    tasks.append(asyncio.create_task(
                        self.run(url_dictionary.copy())))

                elif url_dictionary['download'] is False:
                    print(url)

            elif url_dictionary['download'] is False:
                print(url)

            else:
                url_dictionary['url'] = url
                tasks.append(asyncio.create_task(
                    self.run(url_dictionary.copy())))

        await asyncio.gather(*tasks)



import asyncio
import httpx
import logging
import re

from bs4 import BeautifulSoup
from datetime import datetime
from .common import Common
from ..utils.custom_types import UrlDictionary
from ..utils import color
from ..utils import rat_info

log = logging.getLogger(__name__)

sem = asyncio.Semaphore(6)

# 'https://simp4.jpg.church/1206x2208_bb3d16221cfbf54df75b3d2b6618b5e7.jpg'
re_direct_image_link = re.compile(
    r"(https?://)(\w+\.)?(jpg\.(church|fish))(/|/images/)([\w\-_\(\)\.]+)\.(\w+)")

# 'https://jpg.fish/img/1206x2208-bb3d16221cfbf54df75b3d2b6618b5e7.TmmoSd'
re_image_page = re.compile(
    r"(https?://)(\w+\.)?(jpg\.(church|fish))/img/([\w\-_\(\)\.]+)\.(\w+)")

# 'https://jpg.fish/a/finn.Hu6qp'
re_album_page = re.compile(
    r"(https?://)(\w+\.)?(jpg\.(church|fish))/a/([\w\-_\(\)\.]+)")


class Jpg(Common):

    async def run(self, url_dictionary: UrlDictionary) -> None:

        log.debug("Finding regex match")

        if re_album_page.match(url_dictionary['url']):
            log.debug("Found jpg album")
            await self.album_page(url_dictionary)

        elif re_image_page.match(url_dictionary['url']):
            log.debug("Found jpg image page")
            await self.image_page(url_dictionary)

        elif re_direct_image_link.match(url_dictionary['url']):
            log.debug("Found jpg image link")
            await self.direct_image_link(url_dictionary)

        else:
            log.info(f"No regex match found for link: {url_dictionary['url']}")
            color.output('red', f"url_dictionary['url']")

    async def call(self, url):

        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(url, timeout=None, follow_redirects=True)

        except Exception:
            log.exception(f"Problem getting response: {url}")

        return response

    async def seek_date(self, response: httpx.Response) -> str:

        last_modified_date = datetime.strptime(response.headers.get(
            'last-modified'), "%a, %d %b %Y %H:%M:%S %Z")
        return last_modified_date.strftime('%Y-%m-%d+%H%%3A%M%%3A%S')

    async def direct_image_link(self, url_dictionary: UrlDictionary) -> None:

        async with sem:
            log.debug("Sending request to direct image link")
            response = await self.call(url_dictionary['url'])

            # print(response.headers)
            url_dictionary['response'] = response
            url_dictionary['name'] = re_direct_image_link.match(
                url_dictionary['url']).group(6)
            url_dictionary['extension'] = response.headers.get('content-type')
            url_dictionary['url_to_download'] = url_dictionary['url']
            url_dictionary['file_size'] = response.headers.get(
                'content-length')
            url_dictionary['filename'] = self.common_filename_creator(
                url_dictionary)

            await self.common_file_downloader(url_dictionary.copy())

    async def image_page(self, url_dictionary: UrlDictionary) -> None:

        response = await self.call(url_dictionary['url'])

        soup = BeautifulSoup(response.text, 'html.parser')
        image_container = soup.find('div', {'id': 'image-viewer-container'})
        image_link = image_container.find('img')

        url_dictionary['url'] = image_link['src']
        await self.run(url_dictionary.copy())

    async def album_page(self, url_dictionary: UrlDictionary) -> None:

        url = url_dictionary['url']

        if url[-1] == '/':
            url = url[:-1]

        page = 1
        album_url = f"{url}/?sort=date_desc&page={page}"

        saved_links = []
        while True:

            log.debug(f"Starting pass number {page}")
            log.debug(f"Number of links: {len(saved_links)}")
            response = await self.call(album_url)

            soup = BeautifulSoup(response.text, 'html.parser')
            table = soup.find_all('div', {"class": "list-item"})

            if len(table) == 1:
                for link in table:
                    saved_links.append(link.find('img')['src'])
                log.debug("Found page with only 1 image. Breaking loop")
                log.debug(f"Number of links: {len(saved_links)}")
                break

            for link in table:
                saved_links.append(link.find('img')['src'])

            ids = [id['data-id'] for id in table]
            last_image_id = ids[-1]

            last_image_response = await self.call(saved_links[-1])
            seek_date = await self.seek_date(last_image_response)

            page += 1
            album_url = f"{url}/?sort=date_desc&page={page}&seek={seek_date}.{last_image_id}"

        saved_links = set(saved_links)
        log.debug(f"Total number of links found: {len(saved_links)}")

        tasks = []
        already_downloaded_urls = rat_info.get_downloaded_urls()

        log.debug("Attempting download of untracked links")
        for url in saved_links:
            url = url.replace(".md", "")

            if url in already_downloaded_urls and url_dictionary['download'] is True:
                continue

            elif url not in already_downloaded_urls and url_dictionary['download'] is True:
                url_dictionary['url'] = url
                tasks.append(asyncio.create_task(
                    self.run(url_dictionary.copy())))

            elif url_dictionary['download'] is False:
                print(url)

        await asyncio.gather(*tasks)



import asyncio
import httpx
import json
import logging
import re
import time

from datetime import datetime
from .common import Common
from ..utils.custom_types import UrlDictionary
from ..utils import color
from ..utils import conductor
from ..utils import rat_info

log = logging.getLogger(__name__)

# 'https://i.redd.it/rnjp801xh0ia1.png'
# 'https://v.redd.it/v32qxuvfxxha1/'
re_reddit_media = re.compile(
    r"(https?://)([iv]\.)(reddit|redd\.it)/(\w+)\.?(\w+)?")


# 'https://www.reddit.com/r/ProgrammerHumor/comments/1115ze9/no_one_will_ever_know/'
# 'https://www.reddit.com/r/oddlyterrifying/comments/q2520f/419k_of_you_voted_and_we_listened_this_subreddit/'
re_reddit_post = re.compile(
    r"(https?://)([\w]+\.)(reddit\.com)/(r|user)/(\w+)/(\w+)/(\w{6,7})/(.*)")

# 'https://www.reddit.com/gallery/pyuc2y'
re_reddit_gallery = re.compile(
    r"(https?://)([\w]+\.)?(reddit\.com|redd\.it)/(gallery)/(\w+)")

# 'https://preview.redd.it/ptwafbf4upq71.png?width=1080&crop=smart&auto=webp&v=enabled&s=c1bd0a78e48fcfe976956f14ad19074baea814f3'
re_reddit_gallery_link = re.compile(
    r"(https?://)(preview\.)(redd\.it)/(\w+)\.(\w+)\?(.*)")

# 'https://www.reddit.com/user/GallowBoob/'
re_reddit_user = re.compile(
    r"(https?://)(\w+\.)(reddit\.com)/(user|u)/([\w\-\_]+)(/.*)?")

sem = asyncio.Semaphore(4)


class Reddit(Common):

    def __init__(self):

        self.client_id = "qQ49J_mVAhnh6q4wAoJ8jQ"
        self.user_agent = "Python:ripandtear:0.1.4 (by /u/johnny_barracuda)"
        self.prefix = "reddit"
        self.headers = {}

    async def authorization(self) -> None:

        if self.headers.get("authorization"):
            log.debug("authorization token already found")
            return

        else:
            url = "https://www.reddit.com/api/v1/access_token"

            self.headers['user-agent'] = self.user_agent

            data = {"grant_type": "https://oauth.reddit.com/grants/installed_client",
                    "device_id": "do_not_track_this_device"}

            try:
                log.debug("requestion authorization token")
                async with httpx.AsyncClient() as client:
                    response = await client.post(url, headers=self.headers, data=data,
                                                 auth=(self.client_id, ""), timeout=None)

                data = response.json()
                token = data["access_token"]
                # print(token)
                self.headers["authorization"] = f"bearer {token}"
                log.debug("authorization token set")

            except httpx.ReadError:
                log.error(f"httpx ReadError.")
                return

    async def call(self, endpoint: str, params: dict[str, int]) -> dict:

        url = f"https://oauth.reddit.com/{endpoint}"

        log.debug("getting authorization")
        await self.authorization()
        log.debug("authorization created")

        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(
                    url, params=params, headers=self.headers, timeout=None)

        except httpx.ReadError:
            log.error(f"httpx.ReadError: {url}")
            return

        if response.status_code != 200:

            if response.status_code == 404:

                log.warn(
                    f"Profile was deleted or does not exist - {response.url}")
                color.output('red', str(response.url))
                rat_info.add_entry(category_1='urls_downloaded',
                                   entry=str(response.url))
                return

            if response.status_code == 403:
                log.error(f"Authorization Error - {response.url}")
                color.output('red', str(response.url))
                return

            else:
                log.warn(
                    f"Unable to download {response.status_code}: {response.url}")
            return

        remaining = response.headers.get("x-ratelimit-remaining")

        if remaining and float(remaining) < 2:
            wait_time = int(response.headers["x-ratelimit-reset"])
            log.info(f"Sleeping to reset rateliming: {wait_time} seconds")
            await asyncio.sleep(wait_time)

        data = response.json()

        if "error" in data:

            if data["error"] == 403:
                log.error("Authorization Error")
                return

            if data["error"] == 404:
                log.warn("404 Not Found")
                return

        log.debug("Returning data from call")
        return data

    async def run(self, url_dictionary: UrlDictionary) -> None:

        if re_reddit_media.match(url_dictionary['url']):

            log.debug(
                f"Direct reddit media match: {url_dictionary['url']}")
            await self.reddit_media(url_dictionary.copy())

        elif re_reddit_gallery.match(url_dictionary['url']):

            url_dictionary['reddit_uniq_id'] = re_reddit_gallery.match(
                url_dictionary['url']).group(5)

            log.debug(f"reddit gallery match: {url_dictionary['url']}")
            await self.reddit_post(url_dictionary.copy())

        elif re_reddit_post.match(url_dictionary['url']):

            url_dictionary['reddit_uniq_id'] = re_reddit_post.match(
                url_dictionary['url']).group(7)
            log.debug(f"reddit post match: {url_dictionary['url']}")
            await self.reddit_post(url_dictionary.copy())

        elif re_reddit_user.match(url_dictionary['url']):

            url_dictionary['reddit_username'] = re_reddit_user.match(
                url_dictionary['url']).group(5)
            log.debug(f"reddit user match: {url_dictionary['url']}")
            await self.reddit_user(url_dictionary.copy())

        else:
            log.info(
                f"No matching regex found for {url_dictionary['url']}")

    async def reddit_media(self, url_dictionary: UrlDictionary) -> None:

        log.debug("Inside reddit media download")
        if re_reddit_media.match(url_dictionary['url']).group(2) == 'i.':
            log.debug("reddit image found")

            user_agent = {'user-agent': self.user_agent}

            try:
                async with httpx.AsyncClient() as client:
                    response = await client.get(
                        url_dictionary['url'], headers=user_agent, timeout=None)

                if response.status_code >= 300:
                    log.error(
                        f"Bad status code: {response.status_code} - {response.url}")

                log.debug("200 status code")
                temp = json.dumps(dict(response.headers))
                data = json.loads(temp)
                # print(data)

                url_dictionary['response'] = response
                url_dictionary['prefix'] = self.prefix

                file_creation_date = datetime.strptime(response.headers.get(
                    'last-modified'), "%a, %d %b %Y %H:%M:%S %Z")

                url_dictionary['date'] = time.mktime(
                    file_creation_date.timetuple())

                url_dictionary['name'] = re_reddit_media.match(
                    url_dictionary['url']).group(4)

                url_dictionary['url_to_download'] = str(response.url)

                if data.get('Content-Length'):
                    url_dictionary['file_size'] = int(
                        data['Content-Length'])

                url_dictionary['extension'] = data['content-type']
                url_dictionary['filename'] = self.common_filename_creator(
                    url_dictionary)

                log.debug("url_dictionary sent to downloader")

                await self.common_file_downloader(url_dictionary.copy())

            except Exception:
                log.exception(
                    f"Problem downloading file. Saving for later: {url_dictionary['url_to_download']}")
                color.output('red', f"{url_dictionary['url_to_download']}")
                rat_info.add_error_dictionary(url_dictionary.copy())

        if re_reddit_media.match(url_dictionary['url']).group(2) == 'v.':

            log.debug("Reddit video found")
            async with httpx.AsyncClient() as client:
                response = await client.get(
                    url_dictionary['url'], headers=self.headers, follow_redirects=True, timeout=None)

            # print(response.url)
            url_dictionary['url'] = str(response.url)
            url_dictionary['reddit_uniq_id'] = re_reddit_post.match(
                url_dictionary['url']).group(7)

            await self.reddit_post(url_dictionary.copy())

    async def reddit_post(self, url_dictionary: UrlDictionary) -> None:

        endpoint = f"comments/{url_dictionary['reddit_uniq_id']}/.json"
        params = {}
        params["raw_json"] = "1"

        async with sem:
            log.debug("Calling reddit api for a post")
            data = await self.call(endpoint, params)

            try:
                post = data[0]['data']['children'][0]['data']
                url_dictionary['url_to_download'] = str(post['url'])
                log.info(
                    f"Reddit post found: {url_dictionary['url_to_download']}")

            except TypeError:
                log.warn(
                    f"TypeError while download. Saving in error dictionaries for later: {url_dictionary['reddit_uniq_id']}")
                rat_info.add_error_dictionary(url_dictionary)

            try:
                if conductor.imgur_re.search(url_dictionary['url_to_download']):

                    url_dictionary['url'] = url_dictionary['url_to_download']
                    url_dictionary['url_to_record'] = url_dictionary['url_to_download']
                    url_dictionary['description'] = str(post['title'])
                    log.debug("imgur link found, sending to conductor")
                    await conductor.imgur(url_dictionary.copy())

                elif conductor.redgifs_re.search(url_dictionary['url_to_download']):

                    url_dictionary['url'] = url_dictionary['url_to_download']
                    url_dictionary['description'] = str(post['title'])
                    log.debug("redgifs link found, sending to conductor")
                    await conductor.redgifs(url_dictionary.copy())

                elif conductor.gfycat_re.search(url_dictionary['url_to_download']):

                    url_dictionary['url'] = url_dictionary['url_to_download']
                    url_dictionary['description'] = str(post['title'])
                    log.debug("gfycat link found, sending to conductor")
                    await conductor.gfycat(url_dictionary.copy())

                elif post.get('is_gallery'):

                    if post.get('gallery_data'):
                        await self.reddit_gallery_download(post, url_dictionary)

                    else:
                        log.info(
                            f"Post Deleted: {url_dictionary['reddit_uniq_id']}")
                        color.output(
                            'red', f"https://www.reddit.com/{url_dictionary['reddit_uniq_id']}")
                        rat_info.add_entry(category_1='urls_downloaded',
                                           entry=url_dictionary['url_to_download'])
                        return

                elif re_reddit_media.match(post['url']).group(2) == "i.":

                    url_dictionary['url'] = str(post['url'])
                    url_dictionary['description'] = str(post['title'])
                    url_dictionary['album_name'] = url_dictionary['reddit_uniq_id']

                    await self.reddit_media(url_dictionary.copy())

                elif re_reddit_media.match(post['url']).group(2) == "v.":
                    await self.reddit_video_download(post, url_dictionary.copy())

                else:
                    log.warn(
                        f"No regex match found for: {url_dictionary['url_to_download']}. \
                        Seeing if url matches conductor pattern")
                    await conductor.validate_url(url_dictionary.copy())
                    return

            except AttributeError:

                log.info(
                    f"Not a downloadable link: {url_dictionary['url_to_download']}")
                color.output('red', url_dictionary['url_to_download'])

    async def reddit_user(self, url_dictionary: UrlDictionary) -> None:

        log.info(f"Reddit user found: {url_dictionary['reddit_username']}")
        endpoint = f"user/{url_dictionary['reddit_username']}/submitted/.json"
        params = {}

        tasks = []
        already_downloaded_urls = rat_info.get_downloaded_urls()
        end = False
        while end is False:

            log.debug("Calling Reddit api to get user info")
            params["limit"] = 100
            data = await self.call(endpoint, params)
            log.debug("Received Reddit user data")
            # log.debug(data['data']['children'])

            try:
                for post in data['data']['children']:

                    # if attempting to download the files, skip what has already been downloaded
                    # otherwise if downloading isn't being attempted continue to find the actual urls so they
                    # can be printed (-g flag)

                    if post['data']['url'] in already_downloaded_urls and url_dictionary['download'] is True:
                        continue

                    elif url_dictionary['download'] is False:
                        print(post['data']['url'])

                    elif post['data']['url'] not in already_downloaded_urls and url_dictionary['download'] is True:
                        url_dictionary['reddit_uniq_id'] = post['data']['id']
                        tasks.append(asyncio.create_task(
                            self.reddit_post(url_dictionary.copy())))

                if data['data']['after']:
                    params['after'] = data['data']['after']

                else:
                    end = True

            except TypeError:
                log.info(
                    f"TypeError. Most likely user profile has been deleted. Reddit username: {url_dictionary['reddit_username']}")
                end = True

        await asyncio.gather(*tasks)

    async def reddit_gallery_download(self, post, url_dictionary):

        log.debug("gallery found")
        image_order = []
        for entry in post['gallery_data']['items']:
            id = entry['media_id']
            image_order.append(id)

        tasks = []
        count = 0
        for id in image_order:

            count += 1

            image_link = post['media_metadata'][id]['p'][0]['u']

            url_dictionary['name'] = re_reddit_gallery_link.match(
                image_link).group(4)

            url_dictionary['extension'] = re_reddit_gallery_link.match(
                image_link).group(5)

            image_url = f"https://i.redd.it/{url_dictionary['name']}.{url_dictionary['extension']}"

            async with httpx.AsyncClient() as client:
                response = await client.get(
                    image_url, headers=self.headers, timeout=None)

            temp = json.dumps(dict(response.headers))
            data = json.loads(temp)

            url_dictionary['response'] = response
            url_dictionary['prefix'] = self.prefix
            url_dictionary['date'] = post['created_utc']

            url_dictionary['album_name'] = url_dictionary['reddit_uniq_id']
            url_dictionary['count'] = count
            url_dictionary['description'] = str(post['title'])

            if data.get('content-length'):
                url_dictionary['file_size'] = int(data['content-length'])

            url_dictionary['url_to_download'] = image_url
            url_dictionary['filename'] = self.common_filename_creator(
                url_dictionary.copy())

            log.debug("Sending gallery image to downloader")
            tasks.append(asyncio.create_task(
                self.common_file_downloader(url_dictionary.copy())))

        await asyncio.gather(*tasks)

    async def reddit_video_download(self, post, url_dictionary):

        log.info(f"Reddit video found: {post['url']}")
        url_dictionary['ytdlp_required'] = True
        url_dictionary['prefix'] = self.prefix
        url_dictionary['date'] = post['created_utc']
        url_dictionary['name'] = url_dictionary['reddit_uniq_id']
        url_dictionary['description'] = post['title']
        url_dictionary['url_to_download'] = str(post['url'])
        url_dictionary['extension'] = "mp4"
        url_dictionary['filename'] = self.common_filename_creator(
            url_dictionary.copy())

        await self.common_file_downloader(url_dictionary.copy())



import random
import asyncio
import httpx
import json
import logging
import re
import playwright

from bs4 import BeautifulSoup
from pathlib import Path
from playwright.async_api import async_playwright
from .common import Common
from ..utils.custom_types import UrlDictionary
from ..utils import rat_info
from ..utils import color

log = logging.getLogger(__name__)

# 'https://cdn.bunkr.ru/2316x3088_6f524042147f7b97c8c6e44ccab72374-jvBZFnON.jpg'
# 'https://i.bunkr.ru/2316x3088_6f524042147f7b97c8c6e44ccab72374-jvBZFnON.jpg'
re_bunkr_image_prefix = re.compile(
    r"((i)|([i\w]+)|((cdn\w+)|(cdn)))")

# 'https://media-files10.bunkr.ru/2022-06-22-Natalie-Roush---Leggings-Haul-vJBce0DY.mp4'
re_bunkr_media_files = re.compile(
    r"((media-files[\w]+)|(media-files))")

re_bunkr_image_prefix = re.compile(
    r"((i)|([i\w]+)|((cdn\w+)|(cdn)))")

# 'https://bunkr.su/v/0h1vsyrrw8558lb5agtx4_source-gigzea10.mp4'
re_bunkr_video_href = re.compile(
    r"(/|/[v|d]/)([\w\-_\(\)\.]+)\.(\w+)")

re_bunkr_video_extensions = re.compile(
    r"(mkv|mp4|webm|mov|MOV)")

re_bunkr_file_extensions = re.compile(
    r"(zip|rar|7z|txt)")

# 'https://bunkr.su/a/ouikRoFy'
# 'https://stream.bunkr.ru/v/2022-04-23_Fake-Tan-01-Av9w6Za4.mp4'
re_bunkr_album = re.compile(
    r"(https?://)(bunkr.ru|bunkr.su|bunkr.is)/a/(\w+)")

# 'https://i.bunkr.ru/2316x3088_6f524042147f7b97c8c6e44ccab72374-jvBZFnON.jpg'
re_bunkr_single = re.compile(
    r"(https?://)([\w\-]+\.)?(bunkr\.(ru|la|su|is))(/|/[v|d]/)([\w\-_\(\)\.]+)\.(\w+)")

sem = asyncio.Semaphore(4)
sem_picture = asyncio.Semaphore(6)


class Bunkr(Common):

    async def run(self, url_dictionary: UrlDictionary) -> None:

        if re_bunkr_album.match(url_dictionary['url']):

            try:
                log.debug("Found album. Collecting dictionaries")
                await self.album(url_dictionary)

            except TypeError:
                log.exception("TypeError")
                return

        elif re_bunkr_single.match(url_dictionary['url']):

            log.debug(f"Found single media. {url_dictionary}")
            await self.single(url_dictionary)

        else:
            log.info(
                f"Url doesn't match regex patterns {url_dictionary['url']}")

    async def single(self, url_dictionary: UrlDictionary) -> None:

        try:

            if re_bunkr_media_files.search(url_dictionary['url']):

                log.debug(
                    f"Direct media-files found. {url_dictionary['url']}")
                await self.url_dictionary_builder(url_dictionary.copy())

            elif re_bunkr_file_extensions.match(re_bunkr_single.match(url_dictionary['url']).group(7)):

                log.debug("Direct file found")
                await bunkr_file(url_dictionary['url'])

            elif re_bunkr_single.match(url_dictionary['url']).group(2) == 'stream.':

                log.debug("Direct stream found")
                async with sem:
                    await asyncio.sleep(random.randint(1, 10))
                    await bunkr_stream(url_dictionary)

            elif re_bunkr_video_extensions.match(re_bunkr_single.match(url_dictionary['url']).group(7)):

                log.debug("Video page found. Finding url")
                async with sem:
                    await asyncio.sleep(random.randint(1, 10))
                    await bunkr_stream(url_dictionary)

            elif re_bunkr_image_prefix.match(re_bunkr_single.match(url_dictionary['url']).group(2)):

                if re_bunkr_video_extensions.match(re_bunkr_single.match(url_dictionary['url']).group(7)):

                    log.debug("Direct video found")
                    async with sem:
                        await asyncio.sleep(random.randint(1, 10))
                        await bunkr_stream(url_dictionary)

                else:
                    async with sem_picture:
                        log.debug("Direct image found")
                        await self.url_dictionary_builder(url_dictionary.copy())

            else:
                log.info(
                    f"Given url doesn't match regex pattern {url_dictionary['url']}")
                return

        except AttributeError:
            log.exception(f"Unable to match {url_dictionary['url']}")
            return

    async def url_dictionary_builder(self, url_dictionary: UrlDictionary) -> None:

        try:
            # Bunkr changed domains and where their files are located. Some of them
            # have 'location' in the header to show where the file is currently located
            # while some of them do not. That is why 'location' is check for in the header
            # and if it exists another request is made to the actual location, not just the
            # url that was originally found

            log.debug("Sending request")
            async with httpx.AsyncClient() as client:
                response = await client.get(f"{url_dictionary['url']}", timeout=None, follow_redirects=True)

                log.debug(f"First response headers: {response.headers}")

                if response.headers.get('location'):

                    url = response.headers.get('location')
                    response = await client.get(url, timeout=None)

                    log.debug(f"Second response headers: {response.headers}")

                else:
                    url = response.url

        except httpx.ReadError:
            log.warn(
                f"Unable to confirm - {url_dictionary['url']}. Possible 404")
            return

        # if a video is on a server that is under maintenance, the video is
        # redirected to this link. If this link is found then save the url_dictionary
        # to be downloaded later.

        if response.url == 'https://bnkr.b-cdn.net/maintenance.mp4':
            log.debug(
                f"Server where content is hosted is under maintenance. Saving for later: {url_dictionary['url']}")
            rat_info.add_error_dictionary(url_dictionary)
            return

        temp = json.dumps(dict(response.headers))
        data = json.loads(temp)
        # print(data)

        url_dictionary['response'] = response
        url_dictionary['name'] = re_bunkr_single.match(
            url_dictionary['url']).group(6)

        url_dictionary['url_to_download'] = url

        if data.get('Content-Length'):
            content_length = int(data['Content-Length'])
        else:
            content_length = None

        url_dictionary['file_size'] = content_length
        url_dictionary['extension'] = data['content-type']
        url_dictionary['filename'] = self.common_filename_creator(
            url_dictionary.copy())

        log.debug("Url dictionary built. Sending to downloader")
        await self.common_file_downloader(url_dictionary.copy())

    async def album(self, url_dictionary: UrlDictionary) -> None:

        url = f"{url_dictionary['url']}"

        try:
            async with async_playwright() as pw:
                browser = await pw.chromium.launch(slow_mo=50)
                page = await browser.new_page()
                log.info(f"Navigating to url: {url_dictionary['url']}")
                await page.goto(url, wait_until='networkidle')

                log.info("Searching for content")

                if await page.query_selector("div.grid") is None:

                    if await page.query_selector("div.grid-images"):

                        small_layout = True
                        content = page.locator("div.grid-images > div")

                    else:
                        log.info(
                            f"Page Layout not recognized: {url_dictionary['url']}")
                        return

                else:
                    small_layout = False
                    content = page.locator("div.grid > div")

                for i in range(await content.count()):
                    if i % 50 == 0:
                        log.info(f"Number of content found: {i}")
                    await content.nth(i).hover()

                if small_layout is True:
                    html = await page.locator('div.grid-images').inner_html()

                else:
                    html = await page.locator('div.grid').inner_html()

                await page.close()
                await browser.close()

                soup = BeautifulSoup(html, features='html.parser')

                if len(soup.find_all('a')) > 0:
                    log.info(
                        f"Found {len(soup.find_all('a'))} pieces of content. Preparing for download...")

                    tasks = []
                    for links in soup.find_all('a'):
                        link = str(links.get('href'))
                        if re_bunkr_image_prefix.match(link):
                            url_dictionary['url'] = link
                            tasks.append(asyncio.create_task(
                                self.run(url_dictionary.copy())))
                            continue

                        elif re_bunkr_video_href.match(link):
                            url_dictionary['url'] = f"https://bunkr.su{link}"
                            tasks.append(asyncio.create_task(
                                self.run(url_dictionary.copy())))
                            continue
                        else:
                            log.info(f"Unable to match {link}")

                    await asyncio.gather(*tasks)

                else:
                    log.info("No content found")
                    return

        except Exception:
            log.exception(f"Problem Loading {url_dictionary['url']}")
            await page.close()
            await browser.close()


async def bunkr_stream(url_dictionary: UrlDictionary) -> None:

    async with async_playwright() as pw:
        try:

            browser = await pw.chromium.launch()
            context = await browser.new_context(accept_downloads=True)
            await context.tracing.start(screenshots=True, snapshots=True, sources=True)

            page = await context.new_page()
            response = await page.goto(url_dictionary['url'], wait_until="networkidle")

            if response.status == 404:
                log.warn("404 - Content has been deleted.")
                color.output('red', f"{url_dictionary['url']}")
                await page.close()
                await context.close()
                await browser.close()
                return

            if response.status == 429:
                log.info(
                    f"429 - Too many requests. Saving for later: {url_dictionary['url']}")
                color.output('red', f"{url_dictionary['url']}")
                rat_info.add_error_dictionary(url_dictionary.copy())
                await page.close()
                await context.close()
                await browser.close()
                return

            if response.status >= 500:
                log.info(
                    f"{response.status} - Server error. Saving for later: {url_dictionary['url']}")
                color.output('red', f"{url_dictionary['url']}")
                rat_info.add_error_dictionary(url_dictionary.copy())
                await page.close()
                await context.close()
                await browser.close()
                return

            log.info(f"Searching for video: {url_dictionary['url']}")
            await asyncio.sleep(random.randint(3, 8))

            async with page.expect_download() as download_info:
                await page.locator(
                    "xpath=/html/body/main/section[1]/div/div/div/div[2]/div/a").click()

            download = await download_info.value

            if url_dictionary['download'] is False:
                print(download.url)
                await page.close()
                await context.close()
                await browser.close()
                return

            if rat_info.check_existence(category_1='urls_downloaded', entry=download.url):
                log.info(
                    f"Url already recorded: {download.url}")
                color.output("blue", download.url)
                await page.close()
                await context.close()
                await browser.close()
                return

            filename = download.suggested_filename
            destination_path = Path().cwd() / filename

            log.info(f"Downloading: {filename}")
            print(filename)
            await download.save_as(destination_path)

            log.info(f"Downloaded: {filename}")
            color.output('green', f"{filename}")

            rat_info.add_entry(category_1='urls_downloaded',
                               entry=download.url)

        except playwright._impl._api_types.TimeoutError:
            log.warn(
                f"Page timed out. Server might be under maintenance. Saving for later: {response.status} - {url_dictionary['url']}")
            if url_dictionary['download'] is True:
                color.output("red", f"{url_dictionary['url']}")
            rat_info.add_error_dictionary(url_dictionary.copy())

        finally:
            # await context.tracing.stop(path=f"{random.randint(1, 20)}.zip")
            await page.close()
            await context.close()
            await browser.close()


async def bunkr_file(file_url: str) -> None:

    async with async_playwright() as pw:
        try:
            browser = await pw.chromium.launch()
            context = await browser.new_context(accept_downloads=True)

            page = await context.new_page()
            await page.goto(file_url)

            log.info("Searching for files")
            await page.wait_for_timeout(3000)

            async with page.expect_download() as download_info:
                await page.locator(
                    "xpath=/html/body/main/section[3]/div/div/div/div/div[2]/a").click()

            download = await download_info.value
            filename = download.suggested_filename
            destination_path = Path().cwd() / filename

            log.info(f"Downloading: {filename}")
            print(filename)

            await download.save_as(destination_path)

            log.info(f"Downloaded: {filename}")
            color.output('green', f"{filename}")

            await browser.close()
            rat_info.add_entry(category_1='urls_downloaded',
                               entry=download.url)
        except Exception:
            log.exception("Error downloading file: {file_url}")



import httpx
import logging
import re

from .common import Common
from ..utils import color

log = logging.getLogger(__name__)

re_content_link = re.compile(r"(https?://)(gfycat\.(com)/)(\w+)")

gfycat_api_url = 'https://api.gfycat.com'

prefix = 'gfycat'


class Gfycat(Common):

    async def run(self, url_dictionary):

        if re_content_link.match(url_dictionary['url']):
            content_id = re_content_link.match(url_dictionary['url']).group(4)
            await self.download_individual_gfy(content_id, url_dictionary.copy())

    async def call(self, endpoint):

        api_url = gfycat_api_url + endpoint

        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(api_url, timeout=None)

        except Exception:
            log.exception("Error calling api")

        if response.status_code != 200:
            log.exception(f"Bad status code: {response.status_code}")
            color.output('red', f"{api_url}")

        data = response.json()

        return data

    async def download_individual_gfy(self, content_id, url_dictionary):
        endpoint = f"/v1/gfycats/{content_id}"
        data = await self.call(endpoint)
        url_dictionary['url_to_download'] = data['gfyItem']['mp4Url']
        url_dictionary['prefix'] = prefix
        url_dictionary['name'] = content_id
        url_dictionary['date'] = data['gfyItem']['createDate']
        url_dictionary['file_size'] = data['gfyItem']['mp4Size']
        url_dictionary['extension'] = 'mp4'
        url_dictionary['filename'] = self.common_filename_creator(
            url_dictionary.copy())

        await self.common_file_downloader(url_dictionary.copy())



import json
import logging

from pathlib import Path


log = logging.getLogger(__name__)

rat_template = '''
{
  "names": {
    "reddit": [],
    "redgifs": [],
    "onlyfans": [],
    "fansly": [],
    "pornhub": [],
    "twitter": [],
    "instagram": [],
    "youtube": [],
    "tiktok": [],
    "twitch": [],
    "patreon": [],
    "tumblr": [],
    "myfreecams": [],
    "chaturbate": []
  },
  "links": {
    "coomer": [],
    "simpcity": []
  },
  "urls_to_download": [],
  "urls_downloaded": [],
  "file_hashes": {},
  "error_dictionaries": []
}
'''


def get_rat_name():
    log.debug("Returning .rat name")
    return Path.cwd().stem + '.rat'


def get_rat_template():
    log.debug("Returning copy of rat template")
    return json.loads(rat_template)


def write_rat_file(template):
    log.debug("Writing rat file")
    rat_name = get_rat_name()
    with open(f"{rat_name}", 'w') as f:
        json.dump(template, f, indent=2)


def rat_file_exists():
    log.debug("Checking if .rat file exists")
    files = list(Path.cwd().glob("*.rat"))

    if len(files) > 1:
        log.info("Too many .rat files")
        return False

    elif len(files) == 0:
        # rat_name = get_rat_name()
        # template = json.loads(rat_template)
        log.info("Missing .rat file")
        return False
    elif len(files) == 1:
        return True


def get_rat_info():
    log.debug("Returning stored information from the .rat")

    if rat_file_exists():
        rat_name = get_rat_name()
        with open(f'{rat_name}', 'r') as f:
            rat_contents = json.load(f)
        return rat_contents
    else:
        return 'missing'


def check_existence(category_1, entry):

    rat_contents = get_rat_info()
    if rat_contents == 'too_many' or rat_contents == 'missing':
        pass

    else:

        if category_1 == 'urls_downloaded':
            if entry in rat_contents['urls_downloaded']:
                return True
            else:
                return False


def add_entry(**kwargs):
    if kwargs:
        if kwargs['category_1']:
            category_1 = kwargs['category_1']

        if kwargs['entry']:
            entry = str(kwargs['entry'])

    rat_contents = get_rat_info()
    if rat_contents == 'too_many' or rat_contents == 'missing':
        pass

    else:

        if category_1 == 'urls_downloaded':
            stored_urls = rat_contents['urls_downloaded']
            stored_urls.append(entry)
            rat_contents['urls_downloaded'] = sorted(list(set(stored_urls)))

        rat_name = get_rat_name()
        with open(f"{rat_name}", 'w') as f:
            json.dump(rat_contents, f, indent=2)


def print_rat(category_1, category_2=None, respond=False):

    rat_contents = get_rat_info()

    if rat_contents == 'too_many':
        print("Too many .rat files found")

    elif rat_contents == 'missing':
        print("No .rat file found")

    else:
        try:
            if category_1 == 'names':

                entries: list[str] = []

                for x in rat_contents['names'][category_2]:
                    entries.append(x)

                if respond is True:
                    return entries

                else:
                    print(*entries, sep='\n')

            if category_1 == 'links':
                for x in rat_contents['links'][category_2]:
                    print(x)

            elif category_1 == 'urls_to_download':
                for x in rat_contents['urls_to_download']:
                    print(x)

            elif category_1 == 'urls_downloaded':
                for x in rat_contents['urls_downloaded']:
                    print(x)

            elif category_1 == 'errors':
                for x in rat_contents['error_dictionaries']:
                    print(x)

        except KeyError:
            print("No Entry Found")


def synced_rat_template():

    log.debug("Syncing the current .rat info with the current template")
    rat_contents = get_rat_info()

    if rat_contents == 'too_many':
        log.warn("Too many .rat files found")
        return

    elif rat_contents == 'missing':

        template = get_rat_template()
        return template

    else:

        template = get_rat_template()
        for name in rat_contents['names']:
            if name == 'reddit':
                for y in rat_contents['names']['reddit']:
                    template['names']['reddit'].append(y)

            if name == 'redgifs':
                for y in rat_contents['names']['redgifs']:
                    template['names']['redgifs'].append(y)

            if name == 'onlyfans':
                for y in rat_contents['names']['onlyfans']:
                    template['names']['onlyfans'].append(y)

            if name == 'fansly':
                for y in rat_contents['names']['fansly']:
                    template['names']['fansly'].append(y)

            if name == 'pornhub':
                for y in rat_contents['names']['pornhub']:
                    template['names']['pornhub'].append(y)

            if name == 'twitter':
                for y in rat_contents['names']['twitter']:
                    template['names']['twitter'].append(y)

            if name == 'instagram':
                for y in rat_contents['names']['instagram']:
                    template['names']['instagram'].append(y)

            if name == 'youtube':
                for y in rat_contents['names']['youtube']:
                    template['names']['youtube'].append(y)

            if name == 'tiktok':
                for y in rat_contents['names']['tiktok']:
                    template['names']['tiktok'].append(y)

            if name == 'twitch':
                for y in rat_contents['names']['twitch']:
                    template['names']['twitch'].append(y)

            if name == 'patreon':
                for y in rat_contents['names']['patreon']:
                    template['names']['patreon'].append(y)

            if name == 'tumblr':
                for y in rat_contents['names']['tumblr']:
                    template['names']['tumblr'].append(y)

            if name == 'myfreecams':
                for y in rat_contents['names']['myfreecams']:
                    template['names']['myfreecams'].append(y)

            if name == 'chaturbate':
                for y in rat_contents['names']['chaturbate']:
                    template['names']['chaturbate'].append(y)

        for link in rat_contents['links']:

            if link == 'coomer':
                for link in rat_contents['links']['coomer']:
                    template['links']['coomer'].append(link)

            if link == 'simpcity':
                for link in rat_contents['links']['simpcity']:
                    template['links']['simpcity'].append(link)

        for url in rat_contents['urls_to_download']:
            template['urls_to_download'].append(url)

        for url in rat_contents['urls_downloaded']:
            template['urls_downloaded'].append(url)

        if rat_contents.get('file_hashes'):
            template['file_hashes'] = rat_contents['file_hashes']

        else:
            template['file_hashes'] = {}

        if rat_contents.get('error_dictionaries'):
            template['error_dictionaries'] = rat_contents['error_dictionaries']

        else:
            template['error_dictionaries'] = []

        return template


def update_rat(category_1, category_2=None, category_3=None):

    log.debug("Updating the .rat with new info")
    template = synced_rat_template()

    if category_1 == 'names':
        x = template['names'][category_2]
        x.append(category_3.strip())
        template['names'][category_2] = sorted(list(set(x)))

    if category_1 == 'links':
        x = template['links'][category_2]
        x.append(category_3.strip())
        template['links'][category_2] = sorted(list(set(x)))

    if category_1 == 'urls_to_download':
        x = template['urls_to_download']
        x.append(category_3.strip())
        template['urls_to_download'] = sorted(list(set(x)))

    if category_1 == 'urls_downloaded':
        x = template['urls_downloaded']
        x.append(category_3.strip())
        template['urls_downloaded'] = sorted(list(set(x)))

    if category_1 == 'file_hashes':
        template['file_hashes'] = category_2

    write_rat_file(template)


def get_file_hashes() -> dict[str, str]:

    if rat_file_exists():
        template = synced_rat_template()
        return template['file_hashes']

    else:
        return


def get_urls_to_download() -> list[str] | list:

    if rat_file_exists():
        template = synced_rat_template()
        return template['urls_to_download']

    else:
        return []


def get_downloaded_urls() -> list[str] | list:

    if rat_file_exists():
        template = synced_rat_template()
        return template['urls_downloaded']

    else:
        return []


def add_error_dictionary(url_dictionary):

    template = synced_rat_template()

    if url_dictionary.get('response'):
        del url_dictionary['response']

    error_dictionaries = template['error_dictionaries']
    if len(error_dictionaries) == 0:
        url_dictionary['retries'] = 1
        template['error_dictionaries'] = []
        template['error_dictionaries'].append(url_dictionary)
        write_rat_file(template)

    else:
        already_stored = False
        try:
            for index, error_dictionary in enumerate(template['error_dictionaries']):
                for key, value in error_dictionary.items():
                    if key == 'url' and value == url_dictionary['url']:
                        already_stored = True
                        count = error_dictionary['retries']
                        count += 1
                        log.debug(
                            f"url_dictionary has already been stored. Increasing error attempt count to {count}")
                        if count > 5:
                            log.warn(
                                f"Already attempted 5 times. Removing from being attempted again. {url_dictionary['url_to_download']}")
                            del template['error_dictionaries'][index]
                        else:
                            template['error_dictionaries'][index]['retries'] = count
        except KeyError:
            pass

        if already_stored is False:
            log.debug(
                f"url_dictionary has not had an error yet. Storing it for later")
            url_dictionary['retries'] = 1
            template['error_dictionaries'].append(url_dictionary)

        write_rat_file(template)


def remove_error_dictionary(url_dictionary):

    template = synced_rat_template()

    for index, error_url_dictionary in enumerate(template['error_dictionaries']):
        for key, value in error_url_dictionary:
            if key == 'urls_to_download' and value == url_dictionary['urls_to_download']:
                del template['error_dictionaries'][index]

    write_rat_file(template)


def get_error_dictionaries():

    template = synced_rat_template()

    return template['error_dictionaries']


def erase_error_dictionaries():

    log.debug("Erasing error dictionaries")

    template = synced_rat_template()

    template['error_dictionaries'] = []

    write_rat_file(template)



import logging
import shutil
import re

from pathlib import Path

log = logging.getLogger(__name__)

audio_suffix = re.compile(r"\.(mp3|wav|m4a|)")
gif_suffix = re.compile(r"\.(gif|)")
image_suffix = re.compile(r"\.(jpg|jpeg|png|bmp|webp|)")
text_suffix = re.compile(r"\.(txt|)")
video_suffix = re.compile(r"\.(mp4|webm|mkv|mov|m4v|wmv|avi|flv|mpg|3gp|)")


async def sort():

    files = [file for file in Path().cwd().iterdir() if file.is_file()]

    log.info("Sorting files...")
    try:
        for file in files:

            source = Path(file).name

            if image_suffix.match(Path(file).suffix).group(1):

                destination = (f"pics/{Path(file).name}")
                dir = 'pics'
                await move_file(source, destination, dir)

            elif video_suffix.match(Path(file).suffix).group(1):

                destination = (f"vids/{Path(file).name}")
                dir = 'vids'
                await move_file(source, destination, dir)

            elif gif_suffix.match(Path(file).suffix).group(1):

                destination = (f"gifs/{Path(file).name}")
                dir = 'gifs'
                await move_file(source, destination, dir)

            elif audio_suffix.match(Path(file).suffix).group(1):

                destination = (f"audio/{Path(file).name}")
                dir = 'audio'
                await move_file(source, destination, dir)

            elif text_suffix.match(Path(file).suffix).group(1):

                destination = (f"text/{Path(file).name}")
                dir = 'text'
                await move_file(source, destination, dir)

    except AttributeError:
        log.info(f"Did not find mattching pattern for: {Path(file).name}")


async def file_size(source_file, destination_file):

    source_file_bytes = Path(source_file).stat().st_size

    destination_file_bytes = Path(destination_file).stat().st_size

    return (source_file_bytes, destination_file_bytes)


async def move_file(source, destination, dir):

    if not Path(destination).exists():

        Path(dir).mkdir(exist_ok=True)
        shutil.move(source, dir)

    else:

        file_sizes = await file_size(source, destination)

        if file_sizes[0] > file_sizes[1]:
            Path(destination).unlink()
            shutil.move(source, destination)

        else:
            Path(source).unlink()



import asyncio
import logging
import sys

from .extractors.common import Common
from .utils.custom_types import UrlDictionary
from .utils import cli_arguments
from .utils import conductor
from .utils import logger
from .utils import rat_info
from .utils import file_sorter
from .utils import file_hasher


async def main():

    try:

        parser = cli_arguments.make_args()
        args = parser.parse_args()

        if args.version:
            print("version: ")

        if args.log_level:

            if args.log_level == '1' or args.log_level.lower() == 'debug':
                level = logging.DEBUG
                logger.create_logger(level)
                log = logging.getLogger(__name__)
                log.debug("logging working")

            elif args.log_level == '2' or args.log_level.lower() == 'info':
                level = logging.INFO
                logger.create_logger(level)
                log = logging.getLogger(__name__)
                log.info("logging working")

            elif args.log_level == '3' or args.log_level.lower() == 'warning':
                level = logging.WARNING
                logger.create_logger(level)
                log = logging.getLogger(__name__)
                log.warning("logging working")

            elif args.log_level == '4' or args.log_level.lower() == 'error':
                level = logging.ERROR
                logger.create_logger(level)
                log = logging.getLogger(__name__)
                log.error("logging working")

            elif args.log_level == '5' or args.log_level.lower() == 'critical':
                level = logging.CRITICAL
                logger.create_logger(level)
                log = logging.getLogger(__name__)
                log.critical("logging working")

            else:
                print("Incorrect input to get logging")
        else:
            level = logging.CRITICAL
            logger.create_logger(level)

        if args.print_chaturbate:
            rat_info.print_rat('names', 'chaturbate')

        if args.print_coomer:
            rat_info.print_rat('links', 'coomer')

        if args.print_errors:
            rat_info.print_rat('errors')

        if args.print_fansly:
            rat_info.print_rat('names', 'fansly')

        if args.print_instagram:
            rat_info.print_rat('names', 'instagram')

        if args.print_myfreecams:
            rat_info.print_rat('names', 'myfreecams')

        if args.print_reddit:
            rat_info.print_rat('names', 'reddit')

        if args.print_redgifs:
            rat_info.print_rat('names', 'redgifs')

        if args.print_onlyfans:
            rat_info.print_rat('names', 'onlyfans')

        if args.print_patreon:
            rat_info.print_rat('names', 'patreon')

        if args.print_pornhub:
            rat_info.print_rat('names', 'pornhub')

        if args.print_simpcity:
            rat_info.print_rat('links', 'simpcity')

        if args.print_tiktok:
            rat_info.print_rat('names', 'tiktok')

        if args.print_tumblr:
            rat_info.print_rat('names', 'tumblr')

        if args.print_twitter:
            rat_info.print_rat('names', 'twitter')

        if args.print_twitch:
            rat_info.print_rat('names', 'twitch')

        if args.print_urls_downloaded:
            rat_info.print_rat('urls_downloaded')

        if args.print_urls_to_download:
            rat_info.print_rat('urls_to_download')

        if args.print_youtube:
            rat_info.print_rat('names', 'youtube')

        if args.reddit:
            for entry in args.reddit:
                for name in entry.split(','):
                    rat_info.update_rat('names', 'reddit', name)

        if args.redgifs:
            for entry in args.redgifs:
                for name in entry.split(','):
                    rat_info.update_rat('names', 'redgifs', name)

        if args.onlyfans:
            for entry in args.onlyfans:
                for name in entry.split(','):
                    rat_info.update_rat('names', 'onlyfans', name)

        if args.fansly:
            for entry in args.fansly:
                for name in entry.split(','):
                    rat_info.update_rat('names', 'fansly', name)

        if args.pornhub:
            for entry in args.pornhub:
                for name in entry.split(','):
                    rat_info.update_rat('names', 'pornhub', name)

        if args.twitter:
            for entry in args.twitter:
                for name in entry.split(','):
                    rat_info.update_rat('names', 'twitter', name)

        if args.instagram:
            for entry in args.instagram:
                for name in entry.split(','):
                    rat_info.update_rat('names', 'instagram', name)

        if args.youtube:
            for entry in args.youtube:
                for name in entry.split(','):
                    rat_info.update_rat('names', 'youtube', name)

        if args.tiktok:
            for entry in args.tiktok:
                for name in entry.split(','):
                    rat_info.update_rat('names', 'tiktok', name)

        if args.twitch:
            for entry in args.twitch:
                for name in entry.split(','):
                    rat_info.update_rat('names', 'twitch', name)

        if args.patreon:
            for entry in args.patreon:
                for name in entry.split(','):
                    rat_info.update_rat('names', 'patreon', name)

        if args.tumblr:
            for entry in args.tumblr:
                for name in entry.split(','):
                    rat_info.update_rat('names', 'tumblr', name)

        if args.myfreecams:
            for entry in args.myfreecams:
                for name in entry.split(','):
                    rat_info.update_rat('names', 'myfreecams', name)

        if args.chaturbate:
            for entry in args.chaturbate:
                for name in entry.split(','):
                    rat_info.update_rat('names', 'chaturbate', name)

        if args.simp:
            for entry in args.simp:
                for url in entry.split(','):
                    rat_info.update_rat('links', 'simpcity', url)

        if args.coomer:
            for entry in args.coomer:
                for url in entry.split(','):
                    rat_info.update_rat('links', 'coomer', url)

        if args.url_add:
            for entry in args.url_add:
                for url in entry.split(','):
                    rat_info.update_rat('urls_to_download', None, url)

        if args.urls_downloaded:
            for entry in args.urls_downloaded:
                for url in entry.split(','):
                    rat_info.update_rat('urls_downloaded', None, url)

        if args.erase_errors:
            rat_info.erase_error_dictionaries()

        if args.get_urls:

            await print_urls()

        if args.download:

            await download_urls()

        if args.generic_download:

            for entry in args.generic_download:
                for url in entry.split(','):
                    if args.output:
                        filename: str = args.output.split(',')[0]
                    else:
                        common = Common()
                        await common.common_generic_downloader(url.strip())

        if args.sync_all:
            await sync_all()

        if args.sync_errors:
            await sync_errors()

        if args.sync_reddit:
            await sync_reddit()

        if args.sync_redgifs:
            await sync_redgifs()

        if args.sync_urls_to_download:
            await sync_urls_to_download()

        if args.hash_files:
            await file_hasher.file_hasher()

        if args.sort_files:

            await file_sorter.sort()

    except KeyboardInterrupt:
        print()
        print("Cancelled via keyboard")


def url_dictionary_constructor(url: str, download: bool) -> UrlDictionary:

    url_dictionary: UrlDictionary = {}
    url_dictionary['url'] = url

    url_dictionary['download'] = download

    return url_dictionary


async def download_urls():

    tasks = []

    for entry in args.download:
        for url in entry.split(','):

            url_dictionary: UrlDictionary = url_dictionary_constructor(
                url.strip(), download=True)

            tasks.append(asyncio.create_task(
                conductor.validate_url(url_dictionary.copy())))

    await asyncio.gather(*tasks)


async def print_urls():

    tasks = []

    for entry in args.get_urls:
        for url in entry.split(','):

            url_dictionary: UrlDictionary = url_dictionary_constructor(
                url.strip(), download=False)

            tasks.append(asyncio.create_task(
                conductor.validate_url(url_dictionary.copy())))

    await asyncio.gather(*tasks)


async def sync_reddit() -> None:

    reddit_names: list[str] = rat_info.print_rat(
        'names', 'reddit', respond=True)

    if len(reddit_names) >= 1:

        tasks = []
        for name in reddit_names:
            url: str = f"https://www.reddit.com/user/{name}/submitted"
            url_dictionary: UrlDictionary = url_dictionary_constructor(
                url, download=True)
            tasks.append(asyncio.create_task(
                conductor.validate_url(url_dictionary.copy())))
        await asyncio.gather(*tasks)

    else:
        return


async def sync_redgifs() -> None:

    redgifs_names: list[str] = rat_info.print_rat(
        'names', 'redgifs', respond=True)

    if len(redgifs_names) >= 1:

        tasks = []
        for name in redgifs_names:
            url: str = f"https://www.redgifs.com/users/{name}"
            url_dictionary: UrlDictionary = url_dictionary_constructor(
                url, download=True)
            tasks.append(asyncio.create_task(
                conductor.validate_url(url_dictionary.copy())))
        await asyncio.gather(*tasks)

    else:
        return


async def sync_urls_to_download() -> None:

    urls_to_download = rat_info.get_urls_to_download()

    if len(urls_to_download) >= 1:

        tasks = []
        for url in urls_to_download:
            url_dictionary: UrlDictionary = url_dictionary_constructor(
                url, download=True)
            tasks.append(asyncio.create_task(
                conductor.validate_url(url_dictionary.copy())))
        await asyncio.gather(*tasks)

    else:
        return


async def sync_errors() -> None:

    error_dictionaries = rat_info.get_error_dictionaries()

    if len(error_dictionaries) >= 1:

        tasks = []
        for url_dictionary in error_dictionaries:
            tasks.append(asyncio.create_task(
                conductor.validate_url(url_dictionary.copy())))
        await asyncio.gather(*tasks)

    else:
        return


async def sync_all() -> None:
    await sync_reddit()
    await sync_redgifs()
    await sync_urls_to_download()


def launch():

    try:
        sys.exit(asyncio.run(main()))

    except KeyboardInterrupt:
        sys.exit("\nCancelled via keyboard")

    except RuntimeError:
        pass


if __name__ == "__main__":
    launch()



import asyncio
import httpx
import logging
import re

from bs4 import BeautifulSoup
from .common import Common
from playwright.async_api import async_playwright
from ..utils.custom_types import UrlDictionary
from ..utils import color
from ..utils import rat_info

log = logging.getLogger(__name__)


prefix = "redgifs"
url_root = "https://www.redgifs.com/"
url_root_v3 = "https://v3.redgifs.com/"

single_uid_api = "https://api.redgifs.com/v2/gifs/"

re_extension = re.compile(
    r"(https?://)(\w+\.)(redgifs.com/)(\w+)(-\w+)?\.(\w+)")

re_watch_uid = re.compile(
    r"(/watch/)(\w+)#")

# 'https://i.redgifs.com/i/kosherpurehairstreak.jpg'
# https://v3.redgifs.com/watch/heavyimperfectkookaburra#rel=user%3Alamsinka89
re_single_uid = re.compile(
    r"(https?://)(\w+\.)?(redgifs.com/)(watch|i)/(\w+)")

re_url_to_record = re.compile(
    r"((https?://)(\w+\.)?(redgifs.com/)(watch|i)/(\w+))")

# https://v3.redgifs.com/users/lamsinka89
re_user = re.compile(
    r"((https?://)(\w+\.)?(redgifs\.com)/(users)/(\w+))")

sem = asyncio.Semaphore(6)


class Redgifs(Common):

    async def get_token(self) -> None:

        log.debug("Requesting auth token")
        headers = {'token': ''}

        async with httpx.AsyncClient() as client:
            response = await client.get(
                "https://api.redgifs.com/v2/auth/temporary", headers=headers, timeout=None)

        data = response.json()

        self.headers = {'authorization': f"Bearer {data['token']}"}

        log.debug("Auth token set")

    async def run(self, url_dictionary: UrlDictionary) -> None:

        await self.get_token()

        if re_single_uid.match(url_dictionary['url']):

            log.debug("Single image found. Sending to single single_download")
            await self.single_download(re_single_uid.match(
                url_dictionary['url']).group(5), url_dictionary.copy())

        elif re_user.match(url_dictionary['url']):

            log.debug("User found. Sending to user_download_v3")
            await self.user_download_v3(re_user.match(
                url_dictionary['url']).group(6), url_dictionary.copy())

    async def single_download(self, uid: str, url_dictionary: UrlDictionary) -> None:

        async with sem:

            log.debug("Single image given. Attempting download")

            async with httpx.AsyncClient() as client:
                response = await client.get(
                    f"{single_uid_api}{uid}", headers=self.headers, timeout=None)

            data = response.json()
            # print(data)

            if 'error' in data:

                log.info(f"File Deleted: {url_dictionary['url']}")
                color.output('red', url_dictionary['url'])
                rat_info.add_entry(category_1='urls_downloaded',
                                   entry=url_dictionary['url'])

            else:

                url_dictionary['prefix'] = prefix
                url_dictionary['name'] = data['gif']['id']
                url_dictionary['date'] = data['gif']['createDate']
                url_dictionary['url_to_download'] = data['gif']['urls']['hd']
                url_dictionary['url_to_record'] = re_url_to_record.match(
                    url_dictionary['url']).group(1)

                url_dictionary['file_size'] = None
                url_dictionary['extension'] = re_extension.match(
                    url_dictionary['url_to_download']).group(6)
                url_dictionary['filename'] = self.common_filename_creator(
                    url_dictionary)

                log.debug("url_dictionary built. Sending to downloader")

                await self.common_file_downloader(url_dictionary.copy())

    async def user_download_v3(self, username: str, url_dictionary: UrlDictionary) -> None:

        log.info(f"Redgifs user Found: {username}")

        user_url = f"https://v3.redgifs.com/users/{username}"

        async with async_playwright() as pw:
            browser = await pw.chromium.launch()
            page = await browser.new_page()
            await page.goto(user_url)

            log.info("Searching for videos")
            content = await page.query_selector_all('a.tile')

            stop_scolling = False
            while stop_scolling is False:

                for x in range(5):
                    await page.keyboard.down('PageDown')
                    await page.wait_for_timeout(200)

                new_content = await page.query_selector_all('a.tile')

                if len(new_content) > len(content):

                    content = new_content
                    log.info(f"Gifs found: {len(new_content)}")

                else:

                    content = new_content
                    stop_scolling = True

            html = await page.locator('div.gifList').inner_html()
            await page.close()
            await browser.close()

            soup = BeautifulSoup(html, features='html.parser')

            already_downloaded_urls: list[str] = rat_info.get_downloaded_urls()
            tasks = []
            for links in soup.find_all('a'):

                link = links.get('href')

                uid = re_watch_uid.match(link).group(2)
                url = f"https://v3.redgifs.com/watch/{uid}"

                if url in already_downloaded_urls and url_dictionary['download'] is True:
                    log.info(f"Already downloaded: {url}")
                    continue

                elif url_dictionary['download'] is False:
                    print(url)

                else:
                    log.info(f"Creating task out of {uid}")

                    url_dictionary['url'] = url
                    tasks.append(asyncio.create_task(
                        self.single_download(uid, url_dictionary.copy())))

            log.info("Executing tasks")
            await asyncio.gather(*tasks)



import asyncio
import httpx
import logging
import re
import json

from .common import Common
from ..utils import rat_info
from ..utils import color
from ..utils.custom_types import UrlDictionary

log = logging.getLogger(__name__)

# Errors
# 'https://i.imgur.com/LKvqXWE.gifv'

prefix = 'imgur'

imgur_re_direct_image = re.compile(
    r"(https?://)([im]\.|www\.)?(imgur\.(com|io))/(\w+)(\.jpg|jpeg|png|gifv|mp4|gif)?")
# 'https://i.imgur.com/I7i8u6b.gifv'
# 'https://i.imgur.com/hCOoORb.png'

imgur_re_album = re.compile(
    r"(https?://)?(([im]\.)|(www.))?(imgur\.(com|io))/((a))/(\w+)")
# 'https://imgur.com/a/1GXfC'
# 'https://imgur.com/a/1GXfC/all'

imgur_re_gallery = re.compile(
    r"(https?://)(imgur\.(com|io))/gallery/(\w+)")
# 'https://imgur.com/gallery/mPIxZnC'

headers = {
    "Authorization": "Client-ID 546c25a59c58ad7"
}


class Imgur(Common):

    def imgur_direct_image_hash(self, url: str) -> str:
        log.debug("Looking for direct image hash")
        direct_image_hash = imgur_re_direct_image.search(url).group(5)
        return direct_image_hash

    def imgur_direct_album_hash(self, url: str) -> str:
        log.debug("Looking for album hash")
        direct_album_hash = imgur_re_album.search(url).group(9)
        return direct_album_hash

    def imgur_direct_gallery_hash(self, url: str) -> str:
        log.debug("Looking for gallery hash")
        direct_gallery_hash = imgur_re_gallery.search(url).group(4)
        return direct_gallery_hash

    async def run(self, url_dictionary: UrlDictionary) -> None:

        if imgur_re_gallery.match(url_dictionary['url']):

            log.debug("Gallery found. Sending to gallery download")
            await self.imgur_gallery_download(
                self.imgur_direct_gallery_hash(url_dictionary['url']), url_dictionary.copy())
            # return

        elif imgur_re_album.match(url_dictionary['url']):

            log.debug("Album found. Sending to album download")
            await self.imgur_album_download(
                self.imgur_direct_album_hash(url_dictionary['url']), url_dictionary.copy())
            # return

        elif imgur_re_direct_image.match(url_dictionary['url']):

            log.debug("Direct image found. Sending to single download")
            await self.imgur_single_download(self.imgur_direct_image_hash(
                url_dictionary['url']), url_dictionary.copy())
            # return

        else:
            log.info(f"No regex found for {url_dictionary['url']}")

    async def imgur_single_download(self, image_hash: str, url_dictionary: UrlDictionary) -> None:

        log.info(f"Single image found: {image_hash}")
        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"https://api.imgur.com/3/image/{image_hash}", headers=headers, timeout=None)

        try:

            data = response.json()

        except json.decoder.JSONDecodeError:

            log.warn(f"Image has been deleted: {image_hash}")
            color.output('red', f"\n{url_dictionary['url']}")
            rat_info.add_entry(category_1='urls_downloaded',
                               entry=url_dictionary['url'])
            return

        # print(data)

        # if mp4 is a key in the data dictionary, then it means the image to download
        # ends with .mp4,.gif,.gifv. The mp4 link is grabbed along with its file size.
        # if mp4 does not exist, then the direct link to the file is grabbed along with
        # the file's size

        try:
            if 'mp4' in data['data']:

                url_to_download = data['data']['mp4']
                file_size = data['data']['mp4_size']
                url_dictionary['extension'] = 'mp4'

            else:

                url_to_download = data['data']['link']
                file_size = data['data']['size']
                url_dictionary['extension'] = data['data']['type']

        except KeyError:
            log.warn(f"Url missing keyword. Possibly deleted: {image_hash}")

        url_dictionary['prefix'] = prefix

        # Many times the titles from reddit are better description.
        # If description is true, that means the link was sent here
        # via the reddit ripper along with the better description.
        # use the better description by ignoring what imgur has set
        # other wise use what imgur gives, if availible

        if url_dictionary.get('description'):
            pass

        else:
            url_dictionary['description'] = data['data']['description']

        try:
            url_dictionary['name'] = data['data']['id']
            url_dictionary['date'] = data['data']['datetime']
            url_dictionary['url_to_download'] = url_to_download
            url_dictionary['file_size'] = file_size
            url_dictionary['filename'] = self.common_filename_creator(
                url_dictionary.copy())

            log.info("Sending url_dictionary to downloader")
            await self.common_file_downloader(url_dictionary.copy())

        except KeyError:
            log.warn(f"Url missing keyword. Possibly deleted: {image_hash}")

    async def imgur_album_download(self, album_hash: str, url_dictionary: UrlDictionary) -> None:

        log.debug("Finding album hash")

        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"https://api.imgur.com/3/album/{album_hash}/images", headers=headers, timeout=None)

        try:
            data = response.json()

        except Exception:
            log.info(f"Album deleted: {album_hash}")
            return

        # if the number of images in the album is greater then 1
        # it creates a count variable and passes it on so the pictures
        # will be numbered in the order they are in the album
        # otherwise no numbers are passed on.

        if len(data['data']) > 1:

            log.debug("Multiple images found. Sending them to single download")
            tasks = []
            count = 0
            for entry in data['data']:

                count += 1
                image_hash = self.imgur_direct_image_hash(entry['link'])

                url_dictionary['album_name'] = album_hash
                url_dictionary['count'] = count

                tasks.append(asyncio.create_task(
                    self.imgur_single_download(image_hash, url_dictionary.copy())))

            log.info("Tasks created. Gathering")
            await asyncio.gather(*tasks)

        else:

            for entry in data['data']:
                image_hash = self.imgur_direct_image_hash(entry['link'])
                await self.imgur_single_download(image_hash, url_dictionary.copy())

    async def imgur_gallery_download(self, gallery_hash: str, url_dictionary: UrlDictionary) -> None:

        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(
                    f"https://api.imgur.com/3/gallery/image/{gallery_hash}", headers=headers, timeout=None)

            data = response.json()

            image_hash = self.imgur_direct_image_hash(data['data']['link'])
            log.debug("Image found. Sending to single download")
            await self.imgur_single_download(image_hash, url_dictionary.copy())

        except Exception:

            try:
                log.debug(
                    "Problem loading imgur gallery. Attempting as imgur album instead")
                await self.imgur_album_download(gallery_hash, url_dictionary.copy())

            except Exception:
                log.exception("Problem Downloading a Gallery")



from colorama import Fore, Style


def output(color: str, url: str) -> None:

    # Blue means the url has already been recorded
    if color == 'blue':
        print(Fore.BLUE + f"{url}" + Fore.RESET)
        Style.RESET_ALL

    # Cyan means that the file name alrady exists in the local directory
    if color == 'cyan':
        print(Fore.CYAN + f"{url}" + Fore.RESET)
        Style.RESET_ALL

    # Green means the file is downloading
    if color == 'green':
        print(Fore.GREEN + f"{url}" + Fore.RESET)
        Style.RESET_ALL

    # Red means there was an error with the download. Check logs
    if color == 'red':
        print(Fore.RED + f"{url}" + Fore.RESET)
        Style.RESET_ALL
